"""
Script runs both jobrunner flows in a single process.
"""
import logging
import threading
import time

from opensafely._vendor.jobrunner import config, record_stats, run, sync
from opensafely._vendor.jobrunner.lib.database import ensure_valid_db
from opensafely._vendor.jobrunner.lib.docker import docker
from opensafely._vendor.jobrunner.lib.log_utils import configure_logging
from opensafely._vendor.jobrunner.queries import get_flag_value, set_flag


log = logging.getLogger(__name__)


def start_thread(target, name):
    log.info(f"Starting {name} thread")
    # daemon=True means this thread will be automatically join()ed when the
    # process exits
    thread = threading.Thread(target=target, daemon=True)
    thread.name = name
    thread.start()


def main():
    """Run the main run loop after starting the sync loop in a thread."""
    # extra space to align with other thread's "sync" label.
    threading.current_thread().name = "run "
    fmt = "{asctime} {threadName} {message} {tags}"
    configure_logging(fmt)

    # check db is present and up to date, or else error
    ensure_valid_db()

    try:
        log.info("jobrunner.service started")
        # note: thread name appears in log output, so its nice to keep them all the same length
        start_thread(sync_wrapper, "sync")
        start_thread(record_stats_wrapper, "stat")
        if config.ENABLE_MAINTENANCE_MODE_THREAD:
            start_thread(maintenance_wrapper, "mntn")
        run.main()
    except KeyboardInterrupt:
        log.info("jobrunner.service stopped")


def sync_wrapper():
    """Wrap the sync call with logging context and an exception handler."""
    # avoid busy retries on hard failure
    sleep_after_error = config.POLL_INTERVAL * 5
    while True:
        try:
            sync.main()
        except sync.SyncAPIError as e:
            # Handle these separately as we don't want the full traceback here,
            # just the text of the error response
            log.error(e)
            time.sleep(sleep_after_error)
        except Exception:
            log.exception("Exception in sync thread")
            time.sleep(sleep_after_error)


def record_stats_wrapper():
    """Wrap the record_stats call with logging context and an exception handler."""
    while True:
        try:
            record_stats.main()
            # `main()` should loop forever, if it exits cleanly that means it
            # wasn't configured to run so we should exit the thread rather than
            # looping
            return
        except Exception:
            log.exception("Exception in record_stats thread")
            time.sleep(config.STATS_POLL_INTERVAL)


def maintenance_wrapper():
    """Poll a backend specific way to set the db maintenance flag."""
    while True:
        try:
            maintenance_mode()
        except Exception:
            log.exception("Exception in maintenance_mode thread")

        time.sleep(config.MAINTENANCE_POLL_INTERVAL)


def maintenance_mode():
    """Check if the db is currently in maintenance mode, and set flags as appropriate."""
    # This did not seem big enough to warrant splitting into a separate module.
    log.info("checking if db undergoing maintenance...")
    current = get_flag_value("mode")
    ps = docker(
        [
            "run",
            "--rm",
            "-e",
            "DATABASE_URL",
            "ghcr.io/opensafely-core/cohortextractor",
            "maintenance",
            "--current-mode",
            str(current),
        ],
        env={"DATABASE_URL": config.DATABASE_URLS["full"]},
        check=True,
        capture_output=True,
        text=True,
    )
    last_line = ps.stdout.strip().split("\n")[-1]
    if "db-maintenance" in last_line:
        if current != "db-maintenance":
            log.warning("Enabling DB maintenance mode")
        else:
            log.warning("DB maintenance mode is currently enabled")
        set_flag("mode", "db-maintenance")
    else:
        if current == "db-maintenance":
            log.info("DB maintenance mode had ended")
        set_flag("mode", None)

    mode = get_flag_value("mode")
    log.info(f"db mode: {mode}")
    return mode


if __name__ == "__main__":
    main()



import sys
import json
import subprocess
import opensafely

DESCRIPTION = "Print information about the opensafely tool and available resources"


def add_arguments(parser):
    pass



INFO = """
opensafely version: {version}
docker version: {docker_version}
docker memory: {memory}
docker cpu: {cpu}
"""

def main():
    try:
        ps = subprocess.run(["docker", "info", "-f", "{{json .}}"], capture_output=True, text=True)
        info = json.loads(ps.stdout)
        
        memory = float(info["MemTotal"]) / (1024**3)
        print(INFO.format(
            version=opensafely.__version__,
            docker_version=info["ServerVersion"],
            memory=f"{memory:.1f}G",
            cpu=info["NCPU"],
        ))
    except Exception:
        sys.exit("Error retreiving docker information")



import datetime
import json
import logging
import subprocess
import tempfile
import time
from pathlib import Path

from opensafely._vendor.pipeline.legacy import get_all_output_patterns_from_project_file

from opensafely._vendor.jobrunner import config
from opensafely._vendor.jobrunner.executors.volumes import copy_file, get_volume_api
from opensafely._vendor.jobrunner.job_executor import (
    ExecutorAPI,
    ExecutorState,
    JobDefinition,
    JobResults,
    JobStatus,
    Privacy,
)
from opensafely._vendor.jobrunner.lib import docker
from opensafely._vendor.jobrunner.lib.git import checkout_commit
from opensafely._vendor.jobrunner.lib.path_utils import list_dir_with_ignore_patterns
from opensafely._vendor.jobrunner.lib.string_utils import tabulate


# Directory inside working directory where manifest and logs are created
METADATA_DIR = "metadata"

# Records details of which action created each file
MANIFEST_FILE = "manifest.json"

# This is part of a hack we use to track which files in a volume are newly
# created
TIMESTAMP_REFERENCE_FILE = ".opensafely-timestamp"

# cache of result objects
RESULTS = {}
LABEL = "jobrunner-local"

log = logging.getLogger(__name__)
volume_api = get_volume_api()


def container_name(job):
    return f"os-job-{job.id}"


def get_high_privacy_workspace(workspace):
    return config.HIGH_PRIVACY_WORKSPACES_DIR / workspace


def get_medium_privacy_workspace(workspace):
    if config.MEDIUM_PRIVACY_WORKSPACES_DIR:
        return config.MEDIUM_PRIVACY_WORKSPACES_DIR / workspace
    else:
        return None


def get_log_dir(job):
    # Split log directory up by month to make things slightly more manageable
    month_dir = datetime.date.today().strftime("%Y-%m")
    return config.JOB_LOG_DIR / month_dir / container_name(job)


class LocalDockerError(Exception):
    pass


def get_job_labels(job: JobDefinition):
    """Useful metadata to label docker objects with."""
    return {
        "workspace": job.workspace,
        "action": job.action,
    }


def workspace_is_archived(workspace):
    archive_dir = config.HIGH_PRIVACY_ARCHIVE_DIR
    for ext in config.ARCHIVE_FORMATS:
        path = (archive_dir / workspace).with_suffix(ext)
        if path.exists():
            return True
    return False


class LocalDockerAPI(ExecutorAPI):
    """ExecutorAPI implementation using local docker service."""

    def prepare(self, job):
        current = self.get_status(job)
        if current.state != ExecutorState.UNKNOWN:
            return current

        # Check the workspace is not archived
        workspace_dir = get_high_privacy_workspace(job.workspace)
        if not workspace_dir.exists():
            if workspace_is_archived(job.workspace):
                return JobStatus(
                    ExecutorState.ERROR,
                    f"Workspace {job.workspace} has been archived. Contact the OpenSAFELY tech team to resolve",
                )

        # Check the image exists locally and error if not. Newer versions of
        # docker-cli support `--pull=never` as an argument to `docker run` which
        # would make this simpler, but it looks like it will be a while before this
        # makes it to Docker for Windows:
        # https://github.com/docker/cli/pull/1498
        if not docker.image_exists_locally(job.image):
            log.info(f"Image not found, may need to run: docker pull {job.image}")
            return JobStatus(
                ExecutorState.ERROR,
                f"Docker image {job.image} is not currently available",
            )

        try:
            prepare_job(job)
        except docker.DockerDiskSpaceError as e:
            log.exception(str(e))
            return JobStatus(
                ExecutorState.ERROR, "Out of disk space, please try again later"
            )

        # technically, we're acutally PREPARED, as we did in synchronously, but
        # the loop code is expecting PREPARING, so return that. The next time
        # around the loop, it will pick up that it is PREPARED, and move on.
        return JobStatus(ExecutorState.PREPARING)

    def execute(self, job):
        current = self.get_status(job)
        if current.state != ExecutorState.PREPARED:
            return current

        extra_args = []
        if job.cpu_count:
            extra_args.extend(["--cpus", str(job.cpu_count)])
        if job.memory_limit:
            extra_args.extend(["--memory", job.memory_limit])

        try:
            docker.run(
                container_name(job),
                [job.image] + job.args,
                volume=(volume_api.volume_name(job), "/workspace"),
                env=job.env,
                allow_network_access=job.allow_database_access,
                label=LABEL,
                labels=get_job_labels(job),
                extra_args=extra_args,
            )
        except Exception as exc:
            return JobStatus(
                ExecutorState.ERROR, f"Failed to start docker container: {exc}"
            )

        return JobStatus(ExecutorState.EXECUTING)

    def finalize(self, job):

        current = self.get_status(job)
        if current.state != ExecutorState.EXECUTED:
            return current

        try:
            finalize_job(job)
        except LocalDockerError as exc:
            return JobStatus(ExecutorState.ERROR, f"failed to finalize job: {exc}")

        return JobStatus(ExecutorState.FINALIZING)

    def terminate(self, job):
        docker.kill(container_name(job))
        return JobStatus(ExecutorState.ERROR, "terminated by api")

    def cleanup(self, job):
        if config.CLEAN_UP_DOCKER_OBJECTS:
            log.info("Cleaning up container and volume")
            docker.delete_container(container_name(job))
            volume_api.delete_volume(job)
        else:
            log.info("Leaving container and volume in place for debugging")

        RESULTS.pop(job.id, None)
        return JobStatus(ExecutorState.UNKNOWN)

    def get_status(self, job):
        name = container_name(job)
        job_running = docker.container_inspect(
            name, "State.Running", none_if_not_exists=True
        )

        if job_running is None:
            # no volume for this job found
            if volume_api.volume_exists(job):
                return JobStatus(ExecutorState.PREPARED)
            else:
                return JobStatus(ExecutorState.UNKNOWN)

        elif job_running:
            return JobStatus(ExecutorState.EXECUTING)
        elif job.id in RESULTS:
            return JobStatus(ExecutorState.FINALIZED)
        else:  # container present but not running, i.e. finished
            return JobStatus(ExecutorState.EXECUTED)

    def get_results(self, job):
        if job.id not in RESULTS:
            return JobStatus(ExecutorState.ERROR, "job has not been finalized")

        return RESULTS[job.id]

    def delete_files(self, workspace, privacy, files):
        if privacy == Privacy.HIGH:
            root = get_high_privacy_workspace(workspace)
        elif privacy == Privacy.MEDIUM:
            root = get_medium_privacy_workspace(workspace)
        else:
            raise Exception(f"unknown privacy of {privacy}")

        errors = []
        for name in files:
            path = root / name
            try:
                path.unlink(missing_ok=True)
            except Exception:
                log.exception(f"Could not delete {path}")
                errors.append(name)

        return errors


def prepare_job(job):
    """Creates a volume and populates it with the repo and input files."""
    workspace_dir = get_high_privacy_workspace(job.workspace)

    volume_api.create_volume(job, get_job_labels(job))

    # `docker cp` can't create parent directories for us so we make sure all
    # these directories get created when we copy in the code
    extra_dirs = set(Path(filename).parent for filename in job.inputs)

    try:
        if job.study.git_repo_url and job.study.commit:
            copy_git_commit_to_volume(
                job, job.study.git_repo_url, job.study.commit, extra_dirs
            )
        else:
            # We only encounter jobs without a repo or commit when using the
            # "local_run" command to execute uncommitted local code
            copy_local_workspace_to_volume(job, workspace_dir, extra_dirs)
    except subprocess.CalledProcessError:
        raise LocalDockerError(
            f"Could not checkout commit {job.study.commit} from {job.study.git_repo_url}"
        )

    for filename in job.inputs:
        log.info(f"Copying input file: {filename}")
        if not (workspace_dir / filename).exists():
            raise LocalDockerError(
                f"The file {filename} doesn't exist in workspace {job.workspace} as requested for job {job.id}"
            )
        volume_api.copy_to_volume(job, workspace_dir / filename, filename)

    # Hack: see `get_unmatched_outputs`
    volume_api.touch_file(job, TIMESTAMP_REFERENCE_FILE)


def finalize_job(job):
    container_metadata = docker.container_inspect(
        container_name(job), none_if_not_exists=True
    )
    if not container_metadata:
        raise LocalDockerError("Job container has vanished")
    redact_environment_variables(container_metadata)

    outputs, unmatched_patterns = find_matching_outputs(job)
    unmatched_outputs = get_unmatched_outputs(job, outputs)
    exit_code = container_metadata["State"]["ExitCode"]

    # First get the user-friendly message for known database exit codes, for jobs
    # that have db access
    message = None

    # special case OOMKilled
    if container_metadata["State"]["OOMKilled"]:
        message = "Ran out of memory"
        memory_limit = container_metadata.get("HostConfig", {}).get("Memory", 0)
        if memory_limit > 0:
            gb_limit = memory_limit / (1024**3)
            message += f" (limit for this job was {gb_limit:.2f}GB)"
    else:
        message = config.DOCKER_EXIT_CODES.get(exit_code)

    results = JobResults(
        outputs=outputs,
        unmatched_patterns=unmatched_patterns,
        unmatched_outputs=unmatched_outputs,
        exit_code=container_metadata["State"]["ExitCode"],
        image_id=container_metadata["Image"],
        message=message,
    )
    persist_outputs(job, results.outputs, container_metadata)
    RESULTS[job.id] = results


def persist_outputs(job, outputs, container_metadata):
    """Copy logs and generated outputs to persistant storage."""
    # job_metadata is a big dict capturing everything we know about the state
    # of the job
    job.completed_at = int(time.time())
    job_metadata = dict()
    job_metadata["job_id"] = job.id
    job_metadata["job_request_id"] = job.job_request_id
    job_metadata["created_at"] = job.created_at
    job_metadata["completed_at"] = int(time.time())
    job_metadata["docker_image_id"] = container_metadata["Image"]
    # convert exit code to str so 0 exit codes get logged
    job_metadata["exit_code"] = str(container_metadata["State"]["ExitCode"])
    job_metadata["container_metadata"] = container_metadata
    job_metadata["outputs"] = outputs
    job_metadata["commit"] = job.study.commit

    # Dump useful info in log directory
    log_dir = get_log_dir(job)
    write_log_file(job, job_metadata, log_dir / "logs.txt")
    with open(log_dir / "metadata.json", "w") as f:
        json.dump(job_metadata, f, indent=2)

    # Copy logs to workspace
    workspace_dir = get_high_privacy_workspace(job.workspace)
    metadata_log_file = workspace_dir / METADATA_DIR / f"{job.action}.log"
    copy_file(log_dir / "logs.txt", metadata_log_file)
    log.info(f"Logs written to: {metadata_log_file}")

    # Extract outputs to workspace
    for filename in outputs.keys():
        log.info(f"Extracting output file: {filename}")
        volume_api.copy_from_volume(job, filename, workspace_dir / filename)

    # Copy out logs and medium privacy files
    medium_privacy_dir = get_medium_privacy_workspace(job.workspace)
    if medium_privacy_dir:
        copy_file(
            workspace_dir / METADATA_DIR / f"{job.action}.log",
            medium_privacy_dir / METADATA_DIR / f"{job.action}.log",
        )
        for filename, privacy_level in outputs.items():
            if privacy_level == "moderately_sensitive":
                copy_file(workspace_dir / filename, medium_privacy_dir / filename)

        # this can be removed once osrelease is dead
        write_manifest_file(
            medium_privacy_dir,
            {"repo": job.study.git_repo_url, "workspace": job.workspace},
        )


def find_matching_outputs(job):
    """
    Returns a dict mapping output filenames to their privacy level, plus a list
    of any patterns that had no matches at all
    """
    all_matches = volume_api.glob_volume_files(job)
    unmatched_patterns = []
    outputs = {}
    for pattern, privacy_level in job.output_spec.items():
        filenames = all_matches[pattern]
        if not filenames:
            unmatched_patterns.append(pattern)
        for filename in filenames:
            outputs[filename] = privacy_level
    return outputs, unmatched_patterns


def get_unmatched_outputs(job, outputs):
    """
    Returns all the files created by the job which were *not* matched by any of
    the output patterns.

    This is very useful in debugging because it's easy for users to get their
    output patterns wrong (often just in the wrong directory) and it becomes
    immediately obvious what the problem is if, as well as an error, they can
    see a list of the files the job *did* produce.

    The way we do this is bit hacky, but given that it's only used for
    debugging info and not for Serious Business Purposes, it should be
    sufficient.
    """
    all_outputs = volume_api.find_newer_files(job, TIMESTAMP_REFERENCE_FILE)
    return [filename for filename in all_outputs if filename not in outputs]


def write_log_file(job, job_metadata, filename):
    """
    This dumps the (timestamped) Docker logs for a job to disk, followed by
    some useful metadata about the job and its outputs
    """
    filename.parent.mkdir(parents=True, exist_ok=True)
    docker.write_logs_to_file(container_name(job), filename)
    outputs = sorted(job_metadata["outputs"].items())
    with open(filename, "a") as f:
        f.write("\n\n")
        for key in KEYS_TO_LOG:
            if not job_metadata[key]:
                continue
            f.write(f"{key}: {job_metadata[key]}\n")
        f.write("\noutputs:\n")
        f.write(tabulate(outputs, separator="  - ", indent=2, empty="(no outputs)"))


# Keys of fields to log in manifest.json and log file
KEYS_TO_LOG = [
    "job_id",
    "job_request_id",
    "commit",
    "docker_image_id",
    "exit_code",
    "created_at",
    "completed_at",
]


def copy_git_commit_to_volume(job, repo_url, commit, extra_dirs):
    log.info(f"Copying in code from {repo_url}@{commit}")
    # git-archive will create a tarball on stdout and docker cp will accept a
    # tarball on stdin, so if we wanted to we could do this all without a
    # temporary directory, but not worth it at this stage
    config.TMP_DIR.mkdir(parents=True, exist_ok=True)
    with tempfile.TemporaryDirectory(dir=config.TMP_DIR) as tmpdir:
        tmpdir = Path(tmpdir)
        checkout_commit(repo_url, commit, tmpdir)
        # Because `docker cp` can't create parent directories automatically, we
        # make sure parent directories exist for all the files we're going to
        # copy in later
        for directory in extra_dirs:
            tmpdir.joinpath(directory).mkdir(parents=True, exist_ok=True)
        try:
            volume_api.copy_to_volume(job, tmpdir, ".", timeout=60)
        except docker.DockerTimeoutError:
            # Aborting a `docker cp` into a container at the wrong time can
            # leave the container in a completely broken state where any
            # attempt to interact with or even remove it will just hang, see:
            # https://github.com/docker/for-mac/issues/4491
            #
            # This means we can end up with jobs where any attempt to start
            # them (by copying in code from git) causes the job-runner to
            # completely lock up. To avoid this we use a timeout (60 seconds,
            # which should be more than enough to copy in a few megabytes of
            # code). The exception this triggers will cause the job to fail
            # with an "internal error" message, which will then stop it
            # blocking other jobs. We need a specific exception class here as
            # we need to avoid trying to remove the container, which we would
            # ordinarily do on error, because that operation will also hang :(
            log.exception("Timed out copying code to volume, see issue #154")
            raise LocalDockerError(
                "There was a (hopefully temporary) internal Docker error, "
                "please try the job again"
            )


def copy_local_workspace_to_volume(job, workspace_dir, extra_dirs):
    # To mimic a production run, we only want output files to appear in the
    # volume if they were produced by an explicitly listed dependency. So
    # before copying in the code we get a list of all output patterns in the
    # project and ignore any files matching these patterns
    project_file = workspace_dir / "project.yaml"
    ignore_patterns = get_all_output_patterns_from_project_file(project_file)
    ignore_patterns.extend([".git", METADATA_DIR])
    code_files = list_dir_with_ignore_patterns(workspace_dir, ignore_patterns)

    # Because `docker cp` can't create parent directories automatically, we
    # need to make sure empty parent directories exist for all the files we're
    # going to copy in. For now we do this by actually creating a bunch of
    # empty dirs in a temp directory. It should be possible to do this using
    # the `tarfile` module to talk directly to `docker cp` stdin if we care
    # enough.
    directories = set(Path(filename).parent for filename in code_files)
    directories.update(extra_dirs)
    directories.discard(Path("."))
    if directories:
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            for directory in directories:
                tmpdir.joinpath(directory).mkdir(parents=True, exist_ok=True)
            volume_api.copy_to_volume(job, tmpdir, ".")

    log.info(f"Copying in code from {workspace_dir}")
    for filename in code_files:
        volume_api.copy_to_volume(job, workspace_dir / filename, filename)


# Environment variables whose values do not need to be hidden from the debug
# logs
SAFE_ENVIRONMENT_VARIABLES = set(
    """
    PATH PYTHON_VERSION DEBIAN_FRONTEND DEBCONF_NONINTERACTIVE_SEEN
    UBUNTU_VERSION PYENV_SHELL PYENV_VERSION PYTHONUNBUFFERED
    OPENSAFELY_BACKEND TZ TEMP_DATABASE_NAME PYTHONPATH container LANG LC_ALL
    """.split()
)


def redact_environment_variables(container_metadata):
    """
    Redact the values of any environment variables in the container which
    aren't on the explicit safelist
    """
    env_vars = [line.split("=", 1) for line in container_metadata["Config"]["Env"]]
    redacted_vars = [
        f"{key}=xxxx-REDACTED-xxxx"
        if key not in SAFE_ENVIRONMENT_VARIABLES
        else f"{key}={value}"
        for (key, value) in env_vars
    ]
    container_metadata["Config"]["Env"] = redacted_vars


def write_manifest_file(workspace_dir, manifest):
    manifest_file = workspace_dir / METADATA_DIR / MANIFEST_FILE
    manifest_file_tmp = manifest_file.with_suffix(".tmp")
    manifest_file_tmp.write_text(json.dumps(manifest, indent=2))
    manifest_file_tmp.replace(manifest_file)



import subprocess
import sys
from collections import defaultdict
from http.cookiejar import split_header_words
from pathlib import Path
from urllib.parse import urlparse

from opensafely._vendor import requests
from opensafely._vendor.jobrunner import config
from opensafely._vendor.jobrunner.cli.local_run import docker_preflight_check
from opensafely._vendor.ruamel.yaml import YAML
from opensafely._vendor.ruamel.yaml.error import (YAMLError, YAMLFutureWarning,
                                                  YAMLStreamError, YAMLWarning)

DESCRIPTION = (
    "Command for updating the docker images used to run OpenSAFELY studies locally"
)
REGISTRY = config.DOCKER_REGISTRY
IMAGES = list(config.ALLOWED_IMAGES)
FULL_IMAGES = {f"{REGISTRY}/{image}" for image in IMAGES}
DEPRECATED_REGISTRIES = ["docker.opensafely.org", "ghcr.io/opensafely"]
IMAGES.sort()  # this is just for consistency for testing


def add_arguments(parser):
    choices = ["all"] + IMAGES
    parser.add_argument(
        "image",
        nargs="?",
        choices=choices,
        help="OpenSAFELY docker image to update (default: all)",
        default="all",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Update docker images even if not present locally",
    )
    parser.add_argument(
        "--project",
        help="Use this project to yaml to decide which images to download",
    )


def main(image="all", force=False, project=None):
    if not docker_preflight_check():
        return False

    if project:
        force = True
        images = get_actions_from_project_file(project)
    elif image == "all":
        images = IMAGES
    else:
        # if user has requested a specific image, pull it regardless
        force = True
        images = [image]

    local_images = get_local_images()
    try:
        updated = False
        for image in images:
            # currently databuilder is not published, so we ignore it when pulling
            if image == "databuilder":
                continue
            tag = f"{REGISTRY}/{image}"
            if force or tag in local_images:
                updated = True
                print(f"Updating OpenSAFELY {image} image")
                subprocess.run(["docker", "pull", tag + ":latest"], check=True)

        if updated:
            print("Cleaning up old images")
            remove_deprecated_images(local_images)
            subprocess.run(["docker", "image", "prune", "--force"], check=True)
        else:
            print("No OpenSAFELY docker images found to update.")

    except subprocess.CalledProcessError as exc:
        sys.exit(exc.stderr)


def get_actions_from_project_file(project_yaml):
    path = Path(project_yaml)
    if not path.exists():
        raise RuntimeError(f"Could not find {project_yaml}")

    try:
        with path.open() as f:
            project = YAML(typ="safe", pure=True).load(path)
    except (YAMLError, YAMLStreamError, YAMLWarning, YAMLFutureWarning) as e:
        raise RuntimeError(f"Could not parse {project_yaml}: str(e)")

    images = []
    for action_name, action in project.get("actions", {}).items():
        if not action:
            continue
        command = action.get("run", None)
        if command is None:
            continue

        name, _, version = command.partition(":")
        if name in IMAGES:
            images.append(name)

    if not images:
        raise RuntimeError(f"No actions found in {project_yaml}")

    return images


def get_local_images():
    """Returns a dict of image names and their locally available SHAs"""
    ps = subprocess.run(
        [
            "docker",
            "images",
            "ghcr.io/opensafely-core/*",
            "--no-trunc",
            "--format={{.Repository}}={{.ID}}",
        ],
        check=True,
        text=True,
        capture_output=True,
    )
    images = defaultdict(list)
    for line in ps.stdout.splitlines():
        if not line.strip():
            continue

        name, sha = line.split("=", 1)
        # currently databuilder is not published, so we ignore it
        if "databuilder" in name:
            continue
        images[name].append(sha)

    return images


def remove_deprecated_images(local_images):
    """Temporary clean up functon to remove orphaned images."""
    for registry in DEPRECATED_REGISTRIES:
        for image in IMAGES:
            tag = f"{registry}/{image}"
            if tag in local_images:
                subprocess.run(["docker", "image", "rm", tag], capture_output=True)


session = requests.Session()
token = None


def get_remote_sha(full_name, tag):
    """Get the current sha for a tag from a docker registry."""
    global token

    parsed = urlparse("https://" + full_name)
    manifest_url = f"https://ghcr.io/v2/{parsed.path}/manifests/{tag}"

    if token is None:
        # Docker API requires auth token, even for public resources.
        # However, we can reuse a public token.
        response = session.get(manifest_url)
        token = get_auth_token(response.headers["www-authenticate"])

    response = session.get(manifest_url, headers={"Authorization": f"Bearer {token}"})
    response.raise_for_status()
    return response.json()["config"]["digest"]


def get_auth_token(header):
    """Parse a docker v2 www-authentication header and fetch a token.

    Bearer realm="https://ghcr.io/token",service="ghcr.io",scope="repository:opensafely-core/busybox:pull"
    """
    header = header.lstrip("Bearer")
    # split_header_words is weird, but better than doing it ourselves
    words = split_header_words([header])
    values = dict(next(zip(*words)))
    url = values.pop("realm")
    auth_response = session.get(url, params=values)
    return auth_response.json()["token"]


def check_version():
    need_update = []
    local_images = get_local_images()

    for image in IMAGES:
        full_name = f"{REGISTRY}/{image}"
        local_shas = local_images.get(full_name, [])
        if local_shas:
            latest_sha = get_remote_sha(full_name, "latest")
            if latest_sha not in local_shas:
                need_update.append(image)

    if need_update:
        print(
            f"Warning: the OpenSAFELY docker images for {', '.join(need_update)} actions are out of date - please update by running:\n"
            "    opensafely pull\n"
        )
    return need_update



import argparse
import sys
from pathlib import Path

from opensafely import (
    check, codelists, extract_stats, jupyter, pull, unzip, upgrade, info 
)
from opensafely._vendor.jobrunner.cli import local_run

__version__ = Path(__file__).parent.joinpath("VERSION").read_text().strip()


def main():
    parser = argparse.ArgumentParser()

    def show_help(**kwargs):
        parser.print_help()
        parser.exit()

    parser.set_defaults(function=show_help, handles_unknown_args=False)
    parser.add_argument(
        "--version", action="version", version=f"opensafely {__version__}"
    )

    subparsers = parser.add_subparsers(
        title="available commands", description="", metavar="COMMAND"
    )

    parser_help = subparsers.add_parser("help", help="Show this help message and exit")
    parser_help.set_defaults(function=show_help)

    def add_subcommand(cmd, module):
        subparser = subparsers.add_parser(cmd, help=module.DESCRIPTION)
        subparser.set_defaults(
            handles_unknown_args=False,
            function=module.main,
        )
        module.add_arguments(subparser)

    add_subcommand("run", local_run)
    add_subcommand("codelists", codelists)
    add_subcommand("pull", pull)
    add_subcommand("upgrade", upgrade)
    add_subcommand("check", check)
    add_subcommand("jupyter", jupyter)
    add_subcommand("unzip", unzip)
    add_subcommand("extract-stats", extract_stats)
    add_subcommand("info", info)

    # we version check before parse_args is called so that if a user is
    # following recent documentation but has an old opensafely installed,
    # there's some hint as to why their invocation is failing before being told
    # by argparse.
    if len(sys.argv) == 1 or sys.argv[1] != "upgrade":
        try:
            upgrade.check_version()
        except Exception:
            pass

    if len(sys.argv) == 1 or sys.argv[1] != "pull":
        try:
            pull.check_version()
        except Exception:
            pass

    # start by using looser parsing only known args, so we can get the sub command
    args, unknown = parser.parse_known_args()

    if args.handles_unknown_args:
        kwargs = vars(args)
        kwargs["unknown_args"] = unknown
    # allow supparsers to opt out of looser parser_known_args parsing
    else:
        # reparse args with stricter semantics
        args = parser.parse_args()
        kwargs = vars(args)

    function = kwargs.pop("function")
    kwargs.pop("handles_unknown_args")
    success = function(**kwargs)

    # if `run`ning locally, run `check` in warn mode
    if function == local_run.main and "format-output-for-github" not in kwargs:
        check.main(continue_on_error=True)

    sys.exit(0 if success is not False else 1)


if __name__ == "__main__":
    main()



import sqlite3
import time
from itertools import groupby
from operator import attrgetter

from opensafely._vendor.jobrunner.lib.database import find_all, find_one, find_where, upsert
from opensafely._vendor.jobrunner.models import Flag, Job


def calculate_workspace_state(workspace):
    """
    Return a list containing the most recent uncancelled job (if any) for each action in the workspace. We always
    ignore cancelled jobs when considering the historical state of the system. We also ignore jobs whose action is
    '__error__'; these are dummy jobs created only to help us communicate failure states back to the job-server (see
    create_or_update_jobs.create_failed_job()).
    """
    all_jobs = find_where(Job, workspace=workspace, cancelled=False)
    latest_jobs = []
    for action, jobs in group_by(all_jobs, attrgetter("action")):
        if action == "__error__":
            continue
        ordered_jobs = sorted(jobs, key=attrgetter("created_at"), reverse=True)
        latest_jobs.append(ordered_jobs[0])
    return latest_jobs


def group_by(iterable, key):
    return groupby(sorted(iterable, key=key), key=key)


def get_flag(name):
    """Get a flag from the db"""
    return find_one(Flag, id=name)


def get_flag_value(name, default=None):
    """Get the current value of a flag, with a default"""
    # Note: fail gracefully if the flags table does not exist
    # This means we don't need to worry about it in local_run.
    try:
        return get_flag(name).value
    except ValueError:
        return default
    except sqlite3.OperationalError as e:
        if "no such table" in str(e):
            return default
        raise


def set_flag(name, value, timestamp=None):
    """Set a flag to a value in the db."""
    # Note: table must exist to set flags

    # If it's already in the desired state, do nothing to avoid updating the
    # timestamp
    try:
        current = get_flag(name)
    except ValueError:
        pass
    else:
        if current.value == value:
            return current

    if timestamp is None:
        timestamp = time.time()
    flag = Flag(name, value, timestamp)
    upsert(flag)
    return flag


def get_current_flags():
    """Get all currently set flags"""
    return find_all(Flag)



"""
Custom types to type check the raw input data

When loading data from YAML we get dictionaries which pydantic attempts to
validate.  Some of our validation is done via custom methods using the raw
dictionary data.
"""
from __future__ import annotations

import pathlib
from typing import Any, Dict, TypedDict


RawOutputs = Dict[str, Dict[str, str]]


class RawAction(TypedDict):
    config: dict[Any, Any] | None
    run: str
    needs: list[str] | None
    outputs: RawOutputs
    dummy_data_file: pathlib.Path | None


class RawExpectations(TypedDict):
    population_size: str | int | None


class RawPipeline(TypedDict):
    version: str | float | int
    expectations: RawExpectations
    actions: dict[str, RawAction]



def requires_db_access(args):
    """
    By default actions do not have database access, but certain trusted actions require it
    """
    valid_commands = {
        "cohortextractor": ("generate_cohort", "generate_codelist_report"),
        "databuilder": ("generate-dataset",),
        "sqlrunner": None,  # all commands are valid
    }
    if len(args) <= 1:
        return False

    image, command = args[0], args[1]
    image = image.split(":")[0]
    if image in valid_commands:
        if valid_commands[image] is None or command in valid_commands[image]:
            return True
    return False



import pathlib
import re
import shlex
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Set, TypedDict

from opensafely._vendor.pydantic import BaseModel, root_validator, validator

from .constants import RUN_ALL_COMMAND
from .exceptions import InvalidPatternError
from .features import LATEST_VERSION, get_feature_flags_for_version
from .types import RawOutputs, RawPipeline
from .validation import (
    assert_valid_glob_pattern,
    validate_cohortextractor_outputs,
    validate_databuilder_outputs,
)


cohortextractor_pat = re.compile(r"cohortextractor:\S+ generate_cohort")
databuilder_pat = re.compile(r"databuilder:\S+ generate[-_]dataset")


class Expectations(BaseModel):
    population_size: int

    @validator("population_size", pre=True)
    def validate_population_size(cls, population_size: str) -> int:
        try:
            return int(population_size)
        except (TypeError, ValueError):
            raise ValueError(
                "Project expectations population size must be a number",
            )


class Outputs(BaseModel):
    highly_sensitive: Optional[Dict[str, str]]
    moderately_sensitive: Optional[Dict[str, str]]
    minimally_sensitive: Optional[Dict[str, str]]

    def __len__(self) -> int:
        return len(self.dict(exclude_unset=True))

    @root_validator()
    def at_least_one_output(cls, outputs: Dict[str, str]) -> Dict[str, str]:
        if not any(outputs.values()):
            raise ValueError(
                f"must specify at least one output of: {', '.join(outputs)}"
            )

        return outputs

    @root_validator(pre=True)
    def validate_output_filenames_are_valid(cls, outputs: RawOutputs) -> RawOutputs:
        # we use pre=True here so that we only get the outputs specified in the
        # input data.  With Optional[…] wrapped fields pydantic will set None
        # for us and that just makes the logic a little fiddler with no
        # benefit.
        for privacy_level, output in outputs.items():
            for output_id, filename in output.items():
                try:
                    assert_valid_glob_pattern(filename)
                except InvalidPatternError as e:
                    raise ValueError(f"Output path {filename} is not permitted: {e}")

        return outputs


class Command(BaseModel):
    raw: str  # original string

    class Config:
        # this makes Command hashable, which for some reason due to the
        # Action.parse_run_string works, pydantic requires.
        frozen = True

    @property
    def args(self) -> str:
        return " ".join(self.parts[1:])

    @property
    def name(self) -> str:
        # parts[0] with version split off
        return self.parts[0].split(":")[0]

    @property
    def parts(self) -> List[str]:
        return shlex.split(self.raw)

    @property
    def version(self) -> str:
        # parts[0] with name split off
        return self.parts[0].split(":")[1]


class Action(BaseModel):
    config: Optional[Dict[Any, Any]] = None
    run: Command
    needs: List[str] = []
    outputs: Outputs
    dummy_data_file: Optional[pathlib.Path]

    @validator("run", pre=True)
    def parse_run_string(cls, run: str) -> Command:
        parts = shlex.split(run)

        name, _, version = parts[0].partition(":")
        if not version:
            raise ValueError(
                f"{name} must have a version specified (e.g. {name}:0.5.2)",
            )

        return Command(raw=run)


class PartiallyValidatedPipeline(TypedDict):
    """
    A custom type to type-check the values in "post" root validators

    A root_validator with pre=False (or no kwargs) runs after the values have
    been ingested already, and the `values` arg is a dictionary of model types.

    Note: This is defined here so we don't have to deal with forward reference
    types.
    """

    version: float
    expectations: Expectations
    actions: Dict[str, Action]


class Pipeline(BaseModel):
    version: float
    expectations: Expectations
    actions: Dict[str, Action]

    @property
    def all_actions(self) -> List[str]:
        """
        Get all actions for this Pipeline instance

        We ignore any manually defined run_all action (in later project
        versions this will be an error). We use a list comprehension rather
        than set operators as previously so we preserve the original order.
        """
        return [action for action in self.actions.keys() if action != RUN_ALL_COMMAND]

    @root_validator()
    def validate_actions(
        cls, values: PartiallyValidatedPipeline
    ) -> PartiallyValidatedPipeline:
        # TODO: move to Action when we move name onto it
        validators = {
            cohortextractor_pat: validate_cohortextractor_outputs,
            databuilder_pat: validate_databuilder_outputs,
        }
        for action_id, config in values.get("actions", {}).items():
            for cmd, validator_func in validators.items():
                if cmd.match(config.run.raw):
                    validator_func(action_id, config)

        return values

    @root_validator(pre=True)
    def validate_expectations_per_version(cls, values: RawPipeline) -> RawPipeline:
        """Ensure the expectations key exists for version 3 onwards"""
        try:
            version = float(values["version"])
        except (KeyError, TypeError, ValueError):
            # this is handled in the validate_version_exists and
            # validate_version_value validators
            return values

        feat = get_feature_flags_for_version(version)

        if not feat.EXPECTATIONS_POPULATION:
            # set the default here because pydantic doesn't seem to set it
            # otherwise
            values["expectations"] = {"population_size": 1000}
            return values

        if "expectations" not in values:
            raise ValueError("Project must include `expectations` section")

        if "population_size" not in values["expectations"]:
            raise ValueError(
                "Project `expectations` section must include `population_size` section",
            )

        return values

    @root_validator()
    def validate_outputs_per_version(
        cls, values: PartiallyValidatedPipeline
    ) -> PartiallyValidatedPipeline:
        """
        Ensure outputs are unique for version 2 onwards

        We validate this on Pipeline so we can get the version
        """

        # we're not using pre=True in the validator so we can rely on the
        # version and action keys being the correct type but we have to handle
        # them not existing
        if not (version := values.get("version")):
            return values  # handle missing version

        if (actions := values.get("actions")) is None:
            return values  # hand no actions

        feat = get_feature_flags_for_version(version)
        if not feat.UNIQUE_OUTPUT_PATH:
            return values

        # find duplicate paths defined in the outputs section
        seen_files = []
        for config in actions.values():
            for output in config.outputs.dict(exclude_unset=True).values():
                for filename in output.values():
                    if filename in seen_files:
                        raise ValueError(f"Output path {filename} is not unique")

                    seen_files.append(filename)

        return values

    @root_validator(pre=True)
    def validate_actions_run(cls, values: RawPipeline) -> RawPipeline:
        # TODO: move to Action when we move name onto it
        for action_id, config in values.get("actions", {}).items():
            if config["run"] == "":
                # key is present but empty
                raise ValueError(
                    f"run must have a value, {action_id} has an empty run key"
                )

        return values

    @validator("actions")
    def validate_unique_commands(cls, actions: Dict[str, Action]) -> Dict[str, Action]:
        seen: Dict[Command, List[str]] = defaultdict(list)
        for name, config in actions.items():
            run = config.run
            if run in seen:
                raise ValueError(
                    f"Action {name} has the same 'run' command as other actions: {' ,'.join(seen[run])}"
                )
            seen[run].append(name)

        return actions

    @validator("actions")
    def validate_needs_are_comma_delimited(
        cls, actions: Dict[str, Action]
    ) -> Dict[str, Action]:
        space_delimited = {}
        for name, action in actions.items():
            # find needs definitions with spaces in them
            incorrect = [dep for dep in action.needs if " " in dep]
            if incorrect:
                space_delimited[name] = incorrect

        if not space_delimited:
            return actions

        def iter_incorrect_needs(
            space_delimited: Dict[str, List[str]]
        ) -> Iterable[str]:
            for name, needs in space_delimited.items():
                yield f"Action: {name}"
                for need in needs:
                    yield f" - {need}"

        msg = [
            "`needs` actions should be separated with commas. The following actions need fixing:",
            *iter_incorrect_needs(space_delimited),
        ]

        raise ValueError("\n".join(msg))

    @validator("actions")
    def validate_needs_exist(cls, actions: Dict[str, Action]) -> Dict[str, Action]:
        missing = {}
        for name, action in actions.items():
            unknown_needs = set(action.needs) - set(actions)
            if unknown_needs:
                missing[name] = unknown_needs

        if not missing:
            return actions

        def iter_missing_needs(missing: Dict[str, Set[str]]) -> Iterable[str]:
            for name, needs in missing.items():
                yield f"Action: {name}"
                for need in needs:
                    yield f" - {need}"

        msg = [
            "One or more actions is referencing unknown actions in its needs list:",
            *iter_missing_needs(missing),
        ]
        raise ValueError("\n".join(msg))

    @root_validator(pre=True)
    def validate_version_exists(cls, values: RawPipeline) -> RawPipeline:
        """
        Ensure the version key exists.

        This is a re-implementation of pydantic's field validation so we can
        get a custom error message.  This can be removed when we add a wrapper
        around the models to generate more UI friendly error messages.
        """
        if "version" in values:
            return values

        raise ValueError(
            f"Project file must have a `version` attribute specifying which "
            f"version of the project configuration format it uses (current "
            f"latest version is {LATEST_VERSION})"
        )

    @validator("version", pre=True)
    def validate_version_value(cls, value: str) -> float:
        try:
            return float(value)
        except (TypeError, ValueError):
            raise ValueError(
                f"`version` must be a number between 1 and {LATEST_VERSION}"
            )



import os

from setuptools import find_namespace_packages, setup


with open(os.path.join("opensafely", "VERSION")) as f:
    version = f.read().strip()

setup(
    name="opensafely",
    version=version,
    packages=find_namespace_packages(exclude=["tests"]),
    include_package_data=True,
    url="https://github.com/opensafely-core/opensafely-cli",
    description="Command line tool for running OpenSAFELY studies locally.",
    license="GPLv3",
    author="OpenSAFELY",
    author_email="tech@opensafely.org",
    python_requires=">=3.8",
    entry_points={"console_scripts": ["opensafely=opensafely:main"]},
    classifiers=["License :: OSI Approved :: GNU General Public License v3 (GPLv3)"],
    project_urls={
        "Homepage": "https://opensafely.org",
        "Documentation": "https://docs.opensafely.org",
    },
)



#!/usr/bin/python3
import json
import re
from argparse import ArgumentParser
from collections import Counter
from datetime import datetime
from pathlib import Path

DESCRIPTION = "Parse and output cohortextractor-stats logs as JSONL"
TIMESTAMP_PREFIX_REGEX = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{9}Z")
COMPLETED_PREFIX_REGEX = re.compile(r"^Completed successfully|outputs:")
KEY_VALUE_REGEX = re.compile(r"(?<=\s)([^\s=]+)=(.*?)(?=(?:\s[^\s=]+=|$))")
ACTION_SUMMARY_REGEX = re.compile(
    r"^(state|exit_code|docker_image_id|job_id|job_request_id|run_by_user|created_at|completed_at|local_run):\s(.*)"
)


def add_arguments(parser):
    parser.add_argument(
        "--project-dir",
        help="Project directory (default: current directory)",
        type=Path,
        default=".",
    )
    parser.add_argument(
        "--project-name",
        help="Project name (default: project-dir)",
        type=str,
    )
    parser.add_argument(
        "--output-file",
        help="Output json filename (default: extracted_stats.json)",
        type=Path,
        default="extracted_stats.json",
    )
    return parser


def _timestamp_for_honeytail(ts_datetime):
    honeytail_ts_format = "%Y-%m-%dT%H:%M:%S.000000000Z"
    return ts_datetime.strftime(honeytail_ts_format)


def parse_log(project_name, current_action, job_id, job_request_id, current_log):
    current_log_timestamp = TIMESTAMP_PREFIX_REGEX.match(current_log).group()
    current_log = re.sub(r"\s*\n\s*", " ", current_log).strip()
    log_params = dict(KEY_VALUE_REGEX.findall(current_log))
    return {
        "timestamp": current_log_timestamp,
        "project": project_name,
        "job_id": job_id,
        "job_request_id": job_request_id,
        "action": current_action,
        **log_params,
    }


def parse_stats_logs(raw_logs, project_name, current_action, job_id, job_request_id):
    for log in raw_logs:
        yield parse_log(project_name, current_action, job_id, job_request_id, log)


def _parse_summary_datetime(datetime_string, ts_format):
    """
    Created/completed in the log summary may be a stingified int (unix time)
    or ISO-8061 UTC format
    Convert the string to a datetime object and return both the datetime object and
    its ISO-8061 UTC string
    """
    try:
        datetime_obj = datetime.fromtimestamp(int(datetime_string))
        return datetime_obj, datetime_obj.strftime(ts_format)
    except ValueError:
        return datetime.strptime(datetime_string, ts_format), datetime_string


def format_summary_stats(project_name, current_action, summary_stats):
    start_time = summary_stats.get("created_at")
    end_time = summary_stats.get("completed_at")

    iso_utc_format = "%Y-%m-%dT%H:%M:%SZ"

    start_dt, start_time_formatted = _parse_summary_datetime(start_time, iso_utc_format)
    summary_stats["created_at"] = start_time_formatted
    if start_time and end_time:
        end_dt, end_time_formatted = _parse_summary_datetime(end_time, iso_utc_format)
        elapsed = end_dt - start_dt
        # prefix summary elapsed time to make filtering easier
        summary_stats["action_elapsed_time_secs"] = elapsed.seconds
        summary_stats["action_elapsed_time"] = str(elapsed)
        summary_stats["completed_at"] = end_time_formatted

    return {
        # add the timestamp field in the format that honeytail requires
        "timestamp": _timestamp_for_honeytail(start_dt),
        "project": project_name,
        "action": current_action,
        **{k: v for k, v in summary_stats.items()},
    }


def add_action_counts_by_job_request(summary_stats_list):
    job_request_counts = Counter(
        [
            summary.get("job_request_id")
            for summary in summary_stats_list
            if "job_request_id" in summary
        ]
    )
    # Include the total number of actions run alongside this one in each action summary
    summary_stats_with_action_counts = []
    for summary in summary_stats_list:
        job_request_id = summary.get("job_request_id")
        action_count = job_request_counts[job_request_id] if job_request_id else 0
        summary_stats_with_action_counts.append(
            {**summary, "total_actions_in_job_request": action_count}
        )
    return summary_stats_with_action_counts


def main(project_dir, output_file, project_name=None):
    """
    Read all log files from the project's metadata folder and extract stats logs.

    Cohort-extractor actions log stats with the event "cohortextractor-stats",
    which is used to idenitfy stats logs, and any number of key-value pairs of
    stats data in the format key=value.  Logs can span multiple lines, and values in
    a key=value pair can contain spaces.

    In addition, each log file contains a section with overall job summary data, which
    is extracted into a single additional log, and other non-stats logs, which are ignored.

    Output is in JSONL format, one log per line.
    """
    log_dir = project_dir / "metadata"
    project_name = project_name or project_dir.resolve().name

    log_files = list(log_dir.glob("*.log"))

    stats_logs = []
    summary_stats_logs = []

    for filep in log_files:
        current_action = filep.stem
        summary_stats = {}
        raw_logs = []
        # logs in the log file can span multiple lines, and are a mixture of stats logs,
        # which we want to extract, and other logs, which we mostly don't care about.
        # `extracted_log` keeps track of the log that's being parsed as we interate over the log
        # file line-by-line, collecting the stats logs into the `raw_logs` list

        extracted_log = None
        for line in filep.open().readlines():
            # check if this line is the beginning of a new log or the beginning of the end-of-file
            # action summary
            if TIMESTAMP_PREFIX_REGEX.match(line) or COMPLETED_PREFIX_REGEX.match(line):
                # We're at the beginning of a new log
                # If we were in the process of extracting a log, we're finished with it now
                # Add it to the list of raw logs to be processed later
                # (extracted_log can be None if we're at the beginning of the file, or if
                # the previous line wasn't a stats log)
                if extracted_log is not None:
                    raw_logs.append(extracted_log)
                # Now reset the extracted_log...
                if "cohortextractor-stats" in line:
                    # ...to the current line if it's a stats log
                    extracted_log = line
                else:
                    # ...to None if it's a standard non-stats log
                    extracted_log = None
            elif ACTION_SUMMARY_REGEX.match(line):
                # Check for the summary stats lines
                summary_stats.update(dict(ACTION_SUMMARY_REGEX.findall(line)))
            elif extracted_log is not None:
                # this line isn't a log start, and we're in the process of extracting a log,
                # so it must be a continuation line. Append it to the log we're extracting.
                extracted_log += line
        # Add the final extracted log, if there is one
        if extracted_log is not None:
            raw_logs.append(extracted_log)

        # At the end of each processed file, we parse the raw logs into a dict
        # and add them to the full list that will be written to file
        if raw_logs:
            stats_logs.extend(
                parse_stats_logs(
                    raw_logs,
                    project_name,
                    current_action,
                    summary_stats.get("job_id"),
                    summary_stats.get("job_request_id"),
                )
            )

        # Format and add the summary stats
        # Check for job_id in case we encountered other, non-job logs in the metadata folder
        if summary_stats.get("job_id"):
            summary_stats_logs.append(
                format_summary_stats(project_name, current_action, summary_stats)
            )

    # Now that all log files are processed, find the action acounts by
    # job request
    summary_stats_logs = add_action_counts_by_job_request(summary_stats_logs)

    with (log_dir / output_file).open("w") as outpath:
        for log in [*summary_stats_logs, *stats_logs]:
            json.dump(log, outpath)
            outpath.write("\n")
    print(f"Log stats written to {log_dir / output_file}")


def run():
    parser = ArgumentParser(description=DESCRIPTION)
    parser = add_arguments(parser)
    args = parser.parse_args()
    main(
        project_dir=args.project_dir / "metadata",
        output_file=args.output_file,
        project_name=args.project_name,
    )


if __name__ == "__main__":
    run()



from __future__ import annotations

from .main import load_pipeline


def get_all_output_patterns_from_project_file(project_file: str) -> list[str]:
    config = load_pipeline(project_file)
    all_patterns = set()
    for action in config.actions.values():
        for patterns in action.outputs.dict(exclude_unset=True).values():
            all_patterns.update(patterns.values())
    return list(all_patterns)



from .constants import RUN_ALL_COMMAND
from .exceptions import ProjectValidationError, YAMLError
from .main import load_pipeline


__all__ = [
    "ProjectValidationError",
    "RUN_ALL_COMMAND",
    "YAMLError",
    "load_pipeline",
]



"""
This module is for testing in local development.
"""
import sys

from devtools import debug

from .main import load_pipeline


# just for testing locally
def cli():
    paths = sys.argv[1:]

    if not paths:
        print("No paths given, nothing to do.")
        sys.exit()

    for path in paths:
        with open(path) as f:
            contents = f.read()
        data = load_pipeline(contents, path)
        debug(data)


if __name__ == "__main__":
    cli()
